# investigate the sensitivity of clustering results w.r.t. V (namely, the percentage of variation explained)
import numpy as np
import matplotlib.pyplot as plt
from sklearn.metrics.cluster import adjusted_mutual_info_score, adjusted_rand_score
from ensemble_clustering import functional_data_decomposition, functional_data_clustering

# Create function for percentage of variation and AMI/ARI scores, with eigen_dimension, data, splines, and labels as inputs
def percentage_variation_ami_ari(eigen_dimension, data, spline_length, simulation_labels, K):
    # eigen_dimension = 50
    v_range = np.arange(1, eigen_dimension+1, 1)
    ami_scores = []
    ari_scores = []

    data_smooth, mean, principal_components, eigen_functions = functional_data_decomposition (data, eigen_dimension, spline_length)

    # Investigate the percentage of variation explained by the first 5 eigen dimensions based on the principal components
    variance_explained = np.zeros(eigen_dimension)
    total_variance = 0

    # Determine the total variance
    for i in range(eigen_dimension):
        total_variance += np.var(principal_components[:, i])

    for i in range(eigen_dimension):
        variance_explained[i] = np.var(principal_components[:, i])/total_variance

    print('Percentage of variation explained by each eigen dimensions: ', variance_explained)

    # Determine the ari and ami scores for each eigen dimension
    for i in (v_range):
        # Apply functional data clustering on the principal components until the ith eigen dimension
        membership_matrices, labels = functional_data_clustering(principal_components[:, :i], K)

        ami_scores.append(adjusted_mutual_info_score(simulation_labels, labels))
        ari_scores.append(adjusted_rand_score(simulation_labels, labels))
        
    return variance_explained, ami_scores, ari_scores, v_range

# Function which plots the percentage of variation explained, ami and ari scores against the number of eigen dimensions
def plot_percentage_variation_ami_ari(variance_explained, ari, ami, v_range, simulation_labels, data):

    # Plot the plots on a 2x2 grid
    fig, axs = plt.subplots(2, 2, figsize=(10, 10))
    # Plot the simulated data
    for i in range(len(data)):
        if simulation_labels[i] == 0:
            axs[0, 0].plot(data[i], linestyle='--', color='b')
        elif simulation_labels[i] == 1:
            axs[0, 0].plot(data[i], linestyle='--', color='r')
        elif simulation_labels[i] == 2:
            axs[0, 0].plot(data[i], linestyle='--', color='g')
        elif simulation_labels[i] == 3:
            axs[0, 0].plot(data[i], linestyle='--', color='c')
        elif simulation_labels[i] == 4:
            axs[0, 0].plot(data[i], linestyle='--', color='y')
        elif simulation_labels[i] == 5:
            axs[0, 0].plot(data[i], linestyle='--', color='k')
        else:
            axs[0, 0].plot(data[i], linestyle='--', color='m')
    axs[0, 0].set_title('Simulated data')
    axs[0, 0].set_xlabel('Time')
    axs[0, 0].set_ylabel('Y')
    axs[0, 1].plot(v_range, np.cumsum(variance_explained)*100, linestyle='-', marker='o', color='b')
    axs[0, 1].set_title('Total Variation % '+'Explained per Eigen Dimension')
    axs[0, 1].set_xlabel('Dimension')
    axs[0, 1].set_ylabel('Percentage Explained')
    
    axs[1, 0].plot(v_range, ari, linestyle='-', marker='o', color='b')
    axs[1, 0].set_title('ARI Score Copared to the Eigen Dimension')
    axs[1, 0].set_xlabel('Dimension')
    axs[1, 0].set_ylabel('ARI Score')

    axs[1, 1].plot(v_range, ami, linestyle='-', marker='o', color='b')
    axs[1, 1].set_title('AMI Score Copared to the Eigen Dimension')
    axs[1, 1].set_xlabel('Dimension')
    axs[1, 1].set_ylabel('AMI Score')






from sklearn.decomposition import PCA
from sklearn.mixture import GaussianMixture
from sklearn.cluster import SpectralClustering
import numpy as np
import matplotlib.pyplot as plt
import scipy
from scipy import spatial, interpolate

# First function which takes raw data and V as input arguments, and outputs the smooth functions, population mean, eigen functions and eigen vales
def functional_data_decomposition (data, V, b_spline_length):

    # Smooth the data
    data_smooth = np.zeros((len(data), b_spline_length))
    t = np.expand_dims(np.linspace(0, 10, len(data[0])), 1)

    for i in range(len(data)):
        tck = interpolate.splrep(t, data[i], s=0,k=3) 
        x_new = np.linspace(min(t), max(t), b_spline_length)
        y_fit = interpolate.BSpline(*tck)(x_new)
        y_fit = y_fit.reshape(len(y_fit,))
        data_smooth[i] = y_fit
        # plt.plot(x_new, data_smooth[i], '-m')

    # Get the population mean 
    mean = np.mean(data_smooth, axis=0)

    # plt.plot(x_new, mean, '-k', label='Mean')
    # plt.title((
    #     'Population Functions and Mean'))
    # plt.legend()
    # plt.show()

    # Subtract the mean from the data
    data_smooth_without_mean = data_smooth - mean
    
    # Get principal components of the data
    pca = PCA(n_components=V)
    pca.fit(data_smooth_without_mean)
    principal_componenets = pca.transform(data_smooth_without_mean)

    # Determine the eigen-functions of the population covariance function
    eigen_functions = pca.components_

    # Plot the eigen-functions
    # plt.plot(x_new, eigen_functions.T, '-m')
    # plt.title((
    #     'Eigen-functions of the population'))
    # plt.xlabel('$x$', fontsize=13)
    # plt.ylabel('$y = f(x)$', fontsize=13)
    # plt.show()
    
    return data_smooth, mean, principal_componenets, eigen_functions


# Second function which takes eigen values and clusters K and outputs probabilities and spectral clustering result
def functional_data_clustering (eigen_values, K):
    # Fit a Gaussian mixture model with K components
    membership_matrices = []
    n = len(eigen_values[0])
    for i in range(n):
        gmm = GaussianMixture(n_components=K, covariance_type='full')
        gmm.fit(eigen_values[:, i].reshape(-1, 1))
        labels = gmm.predict(eigen_values[:, i].reshape(-1, 1))
        #Create the membership probabilities for each component
        probability = gmm.predict_proba(eigen_values[:, i].reshape(-1,1))
        membership_matrices.append(probability)

    # Create binary membership matrices for each element in probabilities
    # for i in range(V):
    #     for j in range(n):
    #         for k in range(4):
    #             #Binary membership matrix based on max probability
    #             if probabilities[i][j][k] == max(probabilities[i][j]):
    #                 probabilities[i][j][k] = 1
    #             else:
    #                 probabilities[i][j][k] = 0

    # Affinity matrix A, which is the sum of the V cluster membership matrix M_v*M_v^T
    matrix_multiplication = []
    for i in range(n):
        matrix_multiplication.append(np.matmul(membership_matrices[i], membership_matrices[i].T))
    affinity_matrix = sum(matrix_multiplication)

    # Apply spectral clustering
    sc = SpectralClustering(n_clusters=K, affinity='precomputed', assign_labels='discretize')
    sc.fit(affinity_matrix)
    labels = sc.labels_

    return membership_matrices, labels